import numpy as np
import tensorflow as tf

from .base_tensorflow import BaseTFBackend
from .load_utils import parse_dataset_metadata_bytes, parse_tf_model_bytes
from .predictions import SegmentationPrediction


class Segmenter(BaseTFBackend):
    """Loads a model and uses it to run image segmentation, meaning that it
    takes an image as input and returns a matrix the size of the input image
    where the value of each 'pixel' corresponds to an image segment type.

    Does not support batch prediction TODO: Is this true?
    """

    def __init__(self, model_bytes, metadata_bytes,
                 device: str = None,
                 session_config: tf.ConfigProto = None):
        """
        :param model_bytes: Model file data, likely a loaded *.pb file
        :param metadata_bytes: The dataset metadata file data, likely named
            "dataset_metadata.json"
        :param device: The device to run the model on
        :param session_config: Model configuration options
        """
        super().__init__()

        self.graph, self.session = parse_tf_model_bytes(model_bytes,
                                                        device,
                                                        session_config)
        self.label_map = parse_dataset_metadata_bytes(metadata_bytes)

        # Pull all the necessary inputs from the network
        self.input_image = self.graph.get_tensor_by_name("input")

        # Pull the necessary attributes that we want to get from the network
        # after running it
        self.segmented_tensor = self.graph.get_tensor_by_name(
            "output_prediction")

    def batch_predict(self, img_rgb) -> SegmentationPrediction:
        """Takes numpy BGR images of the format that OpenCV gives, and returns
        predicted labels in the form of a list of DetectionPrediction objects
        from predictions.py.

        :imgs: An OpenCV image
        :param bgr: image is bgr format (bgr=True, default) or rgb (bgr=False)
        :return: List of list of DetectionPrediction objects
        """
        # TODO: The story between RGB and BGR here is confusing
        # This method should take a BGR image and convert it
        # Batch predict for the list of images
        feed = {self.input_image: img_rgb}
        segmented_image = self.session.run(self.segmented_tensor,
                                           feed_dict=feed)

        out = np.squeeze(segmented_image)

        return SegmentationPrediction(out, self.label_map)



from typing import List, Union, Optional

import numpy as np
from vcap.detection_node import DetectionNode

Num = Union[int, float]


def detection_area(detection: DetectionNode):
    bbox = detection.bbox.xywh
    return bbox_area(bbox)


def detection_intersection(detection: DetectionNode,
                           candidates: List[DetectionNode]) -> np.ndarray:
    candidate_bboxes = [candidate.bbox.xywh for candidate in candidates]
    bbox = detection.bbox.xywh
    return bbox_intersection(bbox, candidate_bboxes)


def detection_iou(detection: DetectionNode,
                  candidates: List[DetectionNode]) -> np.ndarray:
    """Computer intersection over union.

    Parameters
    ----------
    bbox : DetectionNode
    candidates : List of DetectionNodes

    Returns
    -------
    ndarray
        The intersection over union in [0, 1] between the `bbox` and each
        candidate. A higher score means a larger fraction of the `bbox` is
        occluded by the candidate.

    """
    candidate_bboxes = [candidate.bbox.xywh for candidate in candidates]
    bbox = detection.bbox.xywh
    return bbox_iou(bbox, candidate_bboxes)


def bbox_area(bbox: List[Num]):
    bbox = np.array(bbox)
    area_bbox = bbox[2:].prod()

    return area_bbox


def bbox_intersection(bbox: List[Num], candidate_bboxes: List[List[Num]]) \
        -> np.ndarray:
    candidate_bboxes = np.array(candidate_bboxes)
    bbox = np.array(bbox)

    bbox_tl, bbox_br = bbox[:2], bbox[:2] + bbox[2:]
    candidates_tl = candidate_bboxes[:, :2]
    candidates_br = candidate_bboxes[:, :2] + candidate_bboxes[:, 2:]

    tl = np.c_[np.maximum(bbox_tl[0], candidates_tl[:, 0])[:, np.newaxis],
               np.maximum(bbox_tl[1], candidates_tl[:, 1])[:, np.newaxis]]
    br = np.c_[np.minimum(bbox_br[0], candidates_br[:, 0])[:, np.newaxis],
               np.minimum(bbox_br[1], candidates_br[:, 1])[:, np.newaxis]]
    wh = np.maximum(0., br - tl)

    area_intersection = wh.prod(axis=1)

    return area_intersection


def bbox_iou(bbox: List[Num], candidate_bboxes: List[List[Num]]) -> np.ndarray:
    """
    :param bbox: Bounding box in the format [x, y, width, height]
    :param candidate_bboxes: A list of bounding boxes in the bbox format
    :return:
    """
    candidate_bboxes = np.array(candidate_bboxes)
    bbox = np.array(bbox)

    area_bbox = bbox_area(bbox)
    area_candidates = candidate_bboxes[:, 2:].prod(axis=1)

    area_intersection = bbox_intersection(bbox, candidate_bboxes)
    area_union = area_bbox + area_candidates - area_intersection

    iou = area_intersection / area_union

    return iou


def iou_cost_matrix(
        detections_a: List[DetectionNode],
        detections_b: List[DetectionNode],
        detections_a_indices: Optional[List[int]] = None,
        detections_b_indices: Optional[List[int]] = None) -> np.ndarray:
    """An intersection over union distance metric.

    Parameters
    ----------
    detections_a : List[vcap.DetectionNode]
        A list of detections.
    detections_b : List[vcap.DetectionNode]
        A list of detections.
    detections_a_indices : Optional[List[int]]
        A list of indices to detections that should be matched. Defaults to
        all `detections`.
    detections_b_indices : Optional[List[int]]
        A list of indices to detections that should be matched. Defaults
        to all `detections`.

    Returns
    -------
    ndarray
        Returns a cost matrix of shape
        len(detections_a_indices), len(detections_b_indices) where entry (i, j) is
        `1 - iou(tracks[track_indices[i]], detections[detection_indices[j]])`.

    """

    if detections_a_indices is None:
        detections_a_indices = np.arange(len(detections_a))
    if detections_b_indices is None:
        detections_b_indices = np.arange(len(detections_b))

    cost_matrix = np.zeros(
        (len(detections_a_indices), len(detections_b_indices)))
    for row, track_idx in enumerate(detections_a_indices):
        detection_from_a = detections_a[track_idx]
        candidate_detections = [detections_b[index]
                                for index in detections_b_indices]
        cost_matrix[row, :] = 1. - detection_iou(detection_from_a,
                                                 candidate_detections)
    return cost_matrix



"""This module is made for running image classification inference easily on
various Slim models
"""
import cv2
import numpy as np
import tensorflow as tf
from collections import namedtuple
from typing import List

from . import load_utils
from .base_tensorflow import BaseTFBackend
from .predictions import ClassificationPrediction


class TFImageClassifier(BaseTFBackend):
    def __init__(self, model_bytes, metadata_bytes, model_name,
                 device: str = None,
                 session_config: tf.ConfigProto = None):
        """
        :param model_bytes: Loaded model data, likely from a *.pb file
        :param metadata_bytes: Loaded dataset metadata, likely from a file
            named "dataset_metadata.json"
        :param model_name: Currently supported model names are listed in the
            model_to_ops_map variable.
        :param device: The device to run the model on
        :param session_config: Model configuration options
        """
        super().__init__()
        assert model_name in model_to_ops_map.keys(), \
            "The model name must be from here: " + str(model_to_ops_map.keys())

        self.config = model_to_ops_map[model_name]
        self.graph, self.session = load_utils.parse_tf_model_bytes(
            model_bytes, device, session_config)
        self.label_map = load_utils.parse_dataset_metadata_bytes(
            metadata_bytes)
        self.class_names = tuple(
            class_name for key, class_name in
            sorted(self.label_map.items(), key=lambda i: i[0]))

        # Create the input node to the graph, with preprocessing built-in
        with self.graph.as_default():
            # Create a new input node for images of various sizes
            self.input_node = tf.placeholder(
                dtype=tf.float32,
                shape=[None, self.config.img_size, self.config.img_size, 3])

            # Create the preprocessing node
            preprocessed = self.config.preprocess(self.input_node)

            # Connect the new input to the loaded graphs input
            self.output_node, = tf.import_graph_def(
                self.graph.as_graph_def(),
                input_map={self.config.input_node: preprocessed},
                return_elements=[self.config.output_node])

        # Tensorflow models are known to process the first batch of frames
        # slowly due to some lazy initialization logic. Send a fake image to
        # force the model to be fully loaded.
        fake_img = np.zeros((50, 50, 3), dtype=np.uint8)
        self.batch_predict([fake_img])

    def batch_predict(self, imgs_bgr) -> List[ClassificationPrediction]:
        """
        :param imgs_bgr: A list of images [img1, img2] in BGR format
        """

        # Preprocess each image
        imgs_bgr_resized = [_resize_and_pad(img, self.config.img_size)
                            for img in imgs_bgr]

        # Batch predict for the list of images
        feed = {self.input_node: imgs_bgr_resized}
        outs = self.session.run(self.output_node,
                                feed_dict=feed)

        # Create the predictions for return
        predictions = [ClassificationPrediction(
            class_scores=out, class_names=self.class_names)
            for out in outs
        ]

        return predictions


def _preprocess_inception_resnet_v2(img):
    # BGR to RGB conversion
    img = img[..., ::-1]

    # Normalize the image between -1 and 1  900-1000 FPS
    img = (2.0 / 255.0) * img - 1.0
    return img


def _preprocess_resnet_v2_50(img):
    # BGR to RGB conversion
    img = img[..., ::-1]
    return img


def _resize_and_pad(img, desired_size):
    """
    Resize an image to the desired width and height
    :param img:
    :param desired_size:
    :return:
    """
    old_size = img.shape[:2]  # old_size is in (height, width) format

    ratio = float(desired_size) / max(old_size)
    new_size = tuple([int(x * ratio) for x in old_size])

    if new_size[0] == 0:
        new_size = (new_size[0] + 1, new_size[1])

    if new_size[1] == 0:
        new_size = (new_size[0], new_size[1] + 1)

    # New_size should be in (width, height) format
    im = cv2.resize(img, (new_size[1], new_size[0]))

    delta_w = desired_size - new_size[1]
    delta_h = desired_size - new_size[0]
    top, bottom = delta_h // 2, delta_h - (delta_h // 2)
    left, right = delta_w // 2, delta_w - (delta_w // 2)

    color = [0, 0, 0]
    img = cv2.copyMakeBorder(im, top, bottom, left, right,
                             cv2.BORDER_CONSTANT,
                             value=color)
    return img


ModelConfig = namedtuple("ModelConfig",
                         ["input_node", "output_node", "img_size",
                          "preprocess"])
resnet_v2_50 = \
    ModelConfig(input_node="resnet_v2_50/Pad:0",
                output_node="resnet_v2_50/predictions/Reshape_1:0",
                img_size=230,
                preprocess=_preprocess_resnet_v2_50)
inception_resnet_v2 = \
    ModelConfig(input_node="input:0",
                output_node="InceptionResnetV2/Logits/Predictions:0",
                img_size=299,
                preprocess=_preprocess_inception_resnet_v2)

inception_v4 = \
    ModelConfig(input_node="input:0",
                output_node="InceptionV4/Logits/Predictions:0",
                img_size=299,
                preprocess=_preprocess_inception_resnet_v2)

mobilenet_v1 = \
    ModelConfig(input_node="input:0",
                output_node="MobilenetV1/Predictions/Reshape_1:0",
                img_size=224,
                preprocess=_preprocess_inception_resnet_v2)

model_to_ops_map = {"resnet_v2_50": resnet_v2_50,
                    "inception_resnet_v2": inception_resnet_v2,
                    "mobilenet_v1": mobilenet_v1,
                    "inception_v4": inception_v4}
"""
models_to_ops_map keeps track of information for different model names, so 
that the Brain interface can work without the user needing to remember these
specific node names.
"""



from itertools import groupby
from typing import List

import numpy as np
import tensorflow as tf

from .base_tensorflow import BaseTFBackend
from .load_utils import parse_dataset_metadata_bytes, parse_tf_model_bytes
from .predictions import DetectionPrediction


class TFObjectDetector(BaseTFBackend):
    """Loads a TensorFlow model and uses it to run object detection, meaning
    that it takes an image as input and returns bounding boxes with
    classifications as output.
    """

    def __init__(self, model_bytes, metadata_bytes,
                 confidence_thresh=0.05,
                 device: str = None,
                 session_config: tf.ConfigProto = None):
        """
        :param model_bytes: Model file data, likely a loaded *.pb file
        :param metadata_bytes: The dataset metadata file data, likely named
            "dataset_metadata.json"
        :param device: The device to run the model on
        :param confidence_thresh: The required confidence threshold to filter
            predictions by. Must be between 0 and 1. There is often an
            additional level of confidence thresholding in the capsule, but
            this is here to avoid creating an unreasonable amount of objects
            for each prediction.
        :param session_config: Model configuration options
        """
        super().__init__()

        assert 0.0 < confidence_thresh < 1.0, \
            "Confidence_thresh must be a number between 0 and 1."

        self.min_confidence = confidence_thresh
        self.graph, self.session = parse_tf_model_bytes(model_bytes,
                                                        device,
                                                        session_config)
        self.label_map = parse_dataset_metadata_bytes(metadata_bytes)

        # Pull all the necessary inputs from the network
        self.image_tensor = self.graph.get_tensor_by_name('image_tensor:0')

        # Pull the necessary attributes that we want to get from the network
        # after running it
        self.boxes_tensor = self.graph.get_tensor_by_name('detection_boxes:0')
        self.scores_tensor = self.graph.get_tensor_by_name('detection_scores:0')
        self.classes_tensor = self.graph.get_tensor_by_name(
            'detection_classes:0')

        # Tensorflow models are known to process the first batch of frames
        # slowly due to some lazy initialization logic. Send a fake image to
        # force the model to be fully loaded.
        fake_img = np.zeros((50, 50, 3), dtype=np.uint8)
        self.batch_predict([fake_img])

    def batch_predict(self, imgs_bgr: List[np.ndarray]) -> List[List[DetectionPrediction]]:
        """Takes a list of numpy BGR images of the format that OpenCV gives, and returns
        predicted labels in the form of a list of DetectionPrediction objects
        from predictions.py

        The code in this function is dedictatd to intelligently sorting
        images by their size, and then batching them. Tensorflow can't batch
        inputs of different sizes, so this is something this function takes
        care of.

        The hardest part is making sure that the outputs are re-ordered back
        to their original input order.

        :imgs_bgr: A list of OpenCV images in BGR format (the default)
        :return: List of list of DetectionPrediction objects
        """
        # TODO: This function desperately needs testing, it's easily broken
        # Preprocess each image
        imgs_rgb = [img[..., ::-1] for img in imgs_bgr]

        # Sort images by their size before batching and running inference
        original_indexes = sorted(range(len(imgs_rgb)),
                                  key=lambda index: imgs_rgb[index].shape)
        sorted_imgs = sorted(imgs_rgb, key=lambda img: img.shape)

        # Run batches of images, grouped by their size, and get the results
        results = []
        for index, group in groupby(sorted_imgs, key=lambda img: img.shape):
            # Run inference on each group of equally sized images
            batch_results = self._process_batch(list(group))
            results += batch_results

        # Reorder the results back to the original order of the images
        reordered_results = sorted(enumerate(results),
                                   key=lambda index_img: original_indexes[
                                       index_img[0]])
        reordered_results = [img for _, img in reordered_results]

        return reordered_results

    def _process_batch(self, imgs_rgb: List[np.ndarray]):
        """This function will ONLY WORK if imgs_bgr contains a list of images
        of the same size (resolution, depth). It returns the results of
        imgs_rgb in their original error."""
        assert all([imgs_rgb[0].shape == img.shape for img in imgs_rgb]), \
            "The resolution/depth of all images in this function must be equal!"

        # Batch predict for the list of images
        feed = {self.image_tensor: imgs_rgb}
        all_boxes, all_scores, all_classes = self.session.run(
            [self.boxes_tensor, self.scores_tensor, self.classes_tensor],
            feed_dict=feed)

        # Convert output to DetectionPrediction format
        all_preds = []
        for img_id in range(len(imgs_rgb)):
            # Format the data to a single long array of boxes, scores, and
            # classes.  Each index corresponds with each other.
            boxes = np.squeeze(all_boxes[img_id])
            scores = np.squeeze(all_scores[img_id])
            classes = np.squeeze(all_classes[img_id])

            labels = self._postprocess_output(imgs_rgb[img_id],
                                              boxes=boxes,
                                              scores=scores,
                                              classes=classes)

            all_preds.append(labels)

        return all_preds

    def _postprocess_output(self, image, boxes, scores, classes):
        # Prepare the prediction output
        h, w, _ = image.shape
        labels = []

        for i, score in enumerate(scores):
            if score < self.min_confidence:
                continue

            # Convert the tensorflow format to 'rect' format of
            # [x1, y1, x2, y2] pixels on frame
            rect = [int(round(boxes[i][1] * w, 0)),
                    int(round(boxes[i][0] * h, 0)),
                    int(round(boxes[i][3] * w, 0)),
                    int(round(boxes[i][2] * h, 0))]

            lbl = DetectionPrediction(
                str(self.label_map[int(classes[i])]),
                classes[i],
                rect,
                scores[i])
            labels.append(lbl)
        return labels



from collections import namedtuple
from typing import List

import cv2
import numpy as np
import tensorflow as tf

from .base_encoder import BaseEncoderBackend
from .base_tensorflow import BaseTFBackend
from .load_utils import parse_tf_model_bytes
from .predictions import EncodingPrediction
from vcap_utils.algorithms import euclidian_distance


class OpenFaceEncoder(BaseEncoderBackend, BaseTFBackend):
    """Encode faces (or whatever the model is capable of encoding)
    using this module. It will easily load and run what is necessary.

    Specifically meant to work with models trained from this repository:
    https://github.com/davidsandberg/facenet
    """

    def __init__(self, model_bytes, model_name,
                 device: str = None,
                 session_config: tf.ConfigProto = None):
        """
        :param model_bytes: Model file bytes, a loaded *.pb file
        :param model_name: The name of the model in order to load correct
        input/output tensor node names
        :param device: The device to run the model on
        :param session_config: Model configuration options
        """
        super().__init__()

        assert model_name in model_to_ops_map.keys(), \
            "The model name must be from here: " + str(model_to_ops_map.keys())

        self.config = model_to_ops_map[model_name]
        self._preprocess = self.config.preprocess
        self.graph, self.session = parse_tf_model_bytes(model_bytes,
                                                        device,
                                                        session_config)

        # Pull all the necessary inputs from the network
        self.image_tensor = self.graph.get_tensor_by_name(
            self.config.input_node)
        self.embedding = self.graph.get_tensor_by_name(
            self.config.embedding)
        self.phase_train_placeholder = self.graph.get_tensor_by_name(
            'phase_train:0')

    def batch_predict(self, imgs_bgr) -> List[EncodingPrediction]:
        imgs_rgb = [self._preprocess(img) for img in imgs_bgr]

        # Batch predict for the list of images
        feed = {self.image_tensor: imgs_rgb,
                self.phase_train_placeholder: False}
        embeddings = self.session.run(self.embedding, feed_dict=feed)
        preds = [EncodingPrediction(e, self.config.distance)
                 for e in embeddings]
        return preds

    def distances(self, encoding_to_compare, encodings):
        return self.config.distance(encoding_to_compare, encodings)


def _preprocess_vggface2_center_loss(img):
    img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
    img = cv2.resize(img, vggface2_center_loss.img_size)

    # Pre-whiten the image
    mean = np.mean(img)
    std = np.std(img)
    std_adj = np.maximum(std, 1.0 / np.sqrt(img.size))
    img = np.multiply(np.subtract(img, mean), 1 / std_adj)

    return img


ModelConfig = namedtuple("ModelConfig",
                         ["input_node", "embedding", "img_size",
                          "preprocess", "distance"])

vggface2_center_loss = ModelConfig(input_node="batch_join:0",
                                   embedding="embeddings:0",
                                   img_size=(160, 160),
                                   preprocess=_preprocess_vggface2_center_loss,
                                   distance=euclidian_distance)
model_to_ops_map = {"vggface2_center_loss": vggface2_center_loss}



from .base_tensorflow import BaseTFBackend
from .base_openvino import BaseOpenVINOBackend
from .tf_object_detection import TFObjectDetector
from .tf_image_classification import TFImageClassifier
from .crowd_density import CrowdDensityCounter
from .depth import DepthPredictor
from .segmentation import Segmenter
from .openface_encoder import OpenFaceEncoder
from .base_encoder import BaseEncoderBackend
from .load_utils import parse_dataset_metadata_bytes, parse_tf_model_bytes
from .predictions import (
    EncodingPrediction,
    SegmentationPrediction,
    ClassificationPrediction,
    DepthPrediction,
    DensityPrediction,
    DetectionPrediction,
)



import numpy as np


def non_max_suppression(detection_nodes, max_bbox_overlap, scores=None):
    """Suppress overlapping detections.

    Original code from [1]_ has been adapted to include confidence score.

    .. [1] http://www.pyimagesearch.com/2015/02/16/
           faster-non-maximum-suppression-python/

    Examples
    --------

        >>> boxes = [d.roi for d in detections]
        >>> scores = [d.confidence for d in detections]
        >>> indices = non_max_suppression(boxes, max_bbox_overlap, scores)
        >>> detections = [detections[i] for i in indices]

    Parameters
    ----------
    boxes : ndarray
        Array of ROIs (x, y, width, height).
    max_bbox_overlap : float
        ROIs that overlap more than this values are suppressed.
    scores : Optional[array_like]
        Detector confidence score.

    Returns
    -------
    List[int]
        Returns indices of detections that have survived non-maxima suppression.

    """
    boxes = np.array([d.bbox.xywh for d in detection_nodes])
    scores = np.array([d.extra_data["detection_confidence"]
                       for d in detection_nodes])

    if len(boxes) == 0:
        return []

    boxes = boxes.astype(np.float)
    pick = []

    x1 = boxes[:, 0]
    y1 = boxes[:, 1]
    x2 = boxes[:, 2] + boxes[:, 0]
    y2 = boxes[:, 3] + boxes[:, 1]

    area = (x2 - x1 + 1) * (y2 - y1 + 1)
    if scores is not None:
        idxs = np.argsort(scores)
    else:
        idxs = np.argsort(y2)

    while len(idxs) > 0:
        last = len(idxs) - 1
        i = idxs[last]
        pick.append(i)

        xx1 = np.maximum(x1[i], x1[idxs[:last]])
        yy1 = np.maximum(y1[i], y1[idxs[:last]])
        xx2 = np.minimum(x2[i], x2[idxs[:last]])
        yy2 = np.minimum(y2[i], y2[idxs[:last]])

        w = np.maximum(0, xx2 - xx1 + 1)
        h = np.maximum(0, yy2 - yy1 + 1)

        overlap = (w * h) / area[idxs[:last]]

        idxs = np.delete(
            idxs, np.concatenate(
                ([last], np.where(overlap > max_bbox_overlap)[0])))

    return [detection_nodes[i] for i in pick]
